from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    ClassVar,
    Generic,
    Iterator,
    Type,
    TypeVar,
    Union,
)

from pydantic import (
    BaseModel,
    Field,
)

from agentlens.exceptions import DatasetIndexError
from agentlens.utils import create_readable_id

MISSING_LABEL = object()
SUBSET_FILTER_FN_INDICATOR = "_is_subset_filter_function"


def Label(
    default: Any = MISSING_LABEL,
    **kwargs,
) -> Field:
    return Field(default=default, **kwargs)


class Row(BaseModel):
    """
    A row in a Dataset. Will not raise validation errors if fields designated
    as Labels are missing
    """

    id: str = Field(default_factory=create_readable_id)

    def __getattribute__(self, name: str) -> Any:
        attr = super().__getattribute__(name)
        if attr is MISSING_LABEL:
            raise AttributeError(f"Missing label: {name}")
        return attr

    # TODO: implement __setattr__ with validation at set-time instead of save-time


RowT = TypeVar("RowT", bound=Row)


class DatasetMetadata(BaseModel):
    version: int
    created_at: datetime
    last_modified: datetime


class DatasetFile(BaseModel, Generic[RowT]):
    metadata: DatasetMetadata
    rows: list[RowT]


class Dataset(Generic[RowT]):
    name: ClassVar[str]
    dataset_dir: ClassVar[Path]
    subset: str | None
    rows: list[RowT]
    row_type: ClassVar[Type[RowT]] = Row
    _file: DatasetFile[RowT]

    def __init__(self, subset: str | None = None):
        self.subset = subset
        self._file = self._read_file()

        rows = self._file.rows
        if subset:
            subset_rows, _ = self.split_rows(subset, rows)
            self.rows = subset_rows
        else:
            self.rows = rows

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        for base in cls.__orig_bases__:
            if getattr(base, "__origin__", None) is Dataset:
                args = base.__args__
                if args:
                    cls.row_type = args[0]
                    break

    @property
    def file_path(self) -> Path:
        # stub = f".{self.subset}" if self.subset else ""
        stub = ""
        path = self.dataset_dir / f"{self.name}{stub}.json"
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def _read_file(self) -> DatasetFile[RowT]:
        if not self.file_path.exists():
            return DatasetFile(
                metadata=DatasetMetadata(
                    version=1,
                    created_at=datetime.now(),
                    last_modified=datetime.now(),
                ),
                rows=[],
            )
        json_str = self.file_path.read_text()
        return DatasetFile[self.row_type].model_validate_json(json_str)

    def split_rows(self, subset: str, rows: list[RowT]) -> tuple[list[RowT], list[RowT]]:
        if not hasattr(self, subset):
            raise ValueError(f"Subset filter '{subset}' not found")

        filter_method = getattr(self, subset)
        if not hasattr(filter_method, SUBSET_FILTER_FN_INDICATOR):
            raise ValueError(f"'{subset}' is not a subset filter")

        subset_rows = []
        other_rows = []

        for row in rows:
            if filter_method(row):
                subset_rows.append(row)
            else:
                other_rows.append(row)

        return subset_rows, other_rows

    def extend(self, rows: list[RowT]) -> None:
        """Appends rows to the dataset in-place"""
        self.rows.extend(rows)

    def clear(self) -> None:
        """Clears the dataset in-place"""
        self.rows.clear()

    def save(self) -> None:
        """
        Saves the current version of the subset to disk, leaving other rows unchanged.
        """
        # concatenate current subset's rows with the rest of the dataset's rows
        new_rows = self.rows
        if self.subset:
            _, other_rows = self.split_rows(self.subset, self._file.rows)
            new_rows.extend(other_rows)

        # overwrite current dataset with new rows
        self._file.rows = new_rows
        json_str = self._file.model_dump_json(indent=2, exclude_defaults=True)
        self.file_path.write_text(json_str)

    def __iter__(self) -> Iterator[RowT]:
        return iter(self.rows)

    def __getitem__(self, key: Union[int, str, Any]) -> RowT:
        if isinstance(key, int):
            if 0 <= key < len(self.rows):
                return self.rows[key]
            raise DatasetIndexError(f"Row index {key} out of range")
        elif isinstance(key, str):
            for row in self:
                if row.id == key:
                    return row
            raise DatasetIndexError(f"Row id '{key}' not found")
        raise TypeError(f"Invalid key type: {type(key)}")

    def __len__(self):
        return len(self.rows)


def subset():
    """Marks a Dataset method as a subset filter"""

    def decorator(func):
        setattr(func, SUBSET_FILTER_FN_INDICATOR, True)
        return func

    return decorator



from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from typing import Callable

from agentlens.dataset import Row


class Hook:
    def __init__(self, cb: Callable, target: Callable, row: Row, **kwargs):
        self.cb = cb
        self.target = target
        self.row = row
        self.kwargs = kwargs

    def __call__(self, output, *args, **kwargs):
        return self.cb(self.row, output, *args, **kwargs, **self.kwargs)


class Hooks:
    _instance: ContextVar[Hooks | None] = ContextVar("hooks_instance", default=None)

    def __init__(self, hooks: list[Hook]):
        self.hooks: dict[str, list[Hook]] = {}
        for hook in hooks:
            self.register(hook)

    @classmethod
    def get(cls) -> Hooks | None:
        return cls._instance.get()

    @classmethod
    @contextmanager
    def create(cls, hooks: list[Hook]):
        instance = cls(hooks)
        token = cls._instance.set(instance)
        try:
            yield instance
        finally:
            cls._instance.reset(token)

    def register(self, hook: Hook) -> None:
        target_name = hook.target.__name__
        if target_name not in self.hooks:
            self.hooks[target_name] = []
        self.hooks[target_name].append(hook)

    def __getitem__(self, func_name: str) -> list[Hook]:
        return self.hooks.get(func_name, [])



from __future__ import annotations

import logging
import os
import uuid
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Protocol,
    TypeVar,
)

import petname
from openai.types import CompletionUsage as OpenAIUsage
from openai.types.chat import (
    ChatCompletion,
)

if TYPE_CHECKING:
    from anthropic.types import Usage as AnthropicUsage

logger = logging.getLogger(__name__)


def create_id():
    return str(uuid.uuid4())


def create_path(path: Path | str) -> Path:
    return path if isinstance(path, Path) else Path(path)


def create_readable_id() -> str:
    return petname.generate(words=4, separator="_")


def join_with_dashes(*args) -> str:
    return "-".join(args)


def merge_args(*args, **kwargs) -> dict:
    return {"args": args, "kwargs": kwargs}


T_Model = TypeVar("T_Model", bound="Response")


class Response(Protocol):
    usage: OpenAIUsage | AnthropicUsage


def update_total_usage(
    response: T_Model | None,
    total_usage: OpenAIUsage | AnthropicUsage,
) -> T_Model | ChatCompletion | None:
    if response is None:
        return None

    response_usage = getattr(response, "usage", None)
    if isinstance(response_usage, OpenAIUsage) and isinstance(total_usage, OpenAIUsage):
        total_usage.completion_tokens += response_usage.completion_tokens or 0
        total_usage.prompt_tokens += response_usage.prompt_tokens or 0
        total_usage.total_tokens += response_usage.total_tokens or 0
        response.usage = total_usage  # Replace each response usage with the total usage
        return response

    # Anthropic usage.
    try:
        from anthropic.types import Usage as AnthropicUsage

        if isinstance(response_usage, AnthropicUsage) and isinstance(total_usage, AnthropicUsage):
            total_usage.input_tokens += response_usage.input_tokens or 0
            total_usage.output_tokens += response_usage.output_tokens or 0
            response.usage = total_usage
            return response
    except ImportError:
        pass

    logger.debug("No compatible response.usage found, token usage not updated.")
    return response


def disable_pydantic_error_url():
    os.environ["PYDANTIC_ERRORS_INCLUDE_URL"] = "0"



import json
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import List, Optional

from pydantic import BaseModel

from agentlens.utils import create_readable_id


class RunLog(BaseModel):
    run_id: str
    status: str
    start_time: str
    end_time: Optional[str] = None


@contextmanager
def create_run_log(run_dir: Path):
    log = RunLogs(run_dir)
    run = log.start_run()
    print("Starting run")
    try:
        yield run
    except Exception:
        run.status = "failed"
        print("Run failed")
        raise
    finally:
        log.finish_run(run)
        if run.status != "failed":
            print("Run completed")


class RunLogs:
    def __init__(self, run_dir: Path):
        self._run_dir = run_dir
        self.runs = self.load()

    def file_path(self) -> Path:
        path = self._run_dir / "run_log.json"
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def load(self) -> List[RunLog]:
        path = self.file_path()
        if not path.exists():
            return []
        with open(path) as f:
            return [RunLog(**entry) for entry in json.load(f)]

    def save(self):
        with open(self.file_path(), "w") as f:
            json.dump([run.model_dump() for run in self.runs], f, indent=2)

    def create_id(self) -> str:
        counter = 0
        while True and counter < 100:
            id = create_readable_id()
            if id not in {run.run_id for run in self.runs}:
                return id
            counter += 1  # just in case
        raise Exception("Failed to create a unique run id")

    def start_run(self) -> RunLog:
        run = RunLog(
            run_id=self.create_id(),
            status="started",
            start_time=datetime.now().isoformat(),
        )
        self.runs.append(run)
        self.save()
        return run

    def finish_run(self, run: RunLog):
        if run.status != "failed":  # Only update status if not already failed
            run.status = "completed"
        run.end_time = datetime.now().isoformat()
        self.save()



import asyncio
import logging
import random
from functools import wraps
from pathlib import Path
from typing import Any, Awaitable, Callable, ParamSpec, Sequence, Type, TypeVar

import nest_asyncio
from langfuse import Langfuse
from langfuse.decorators import langfuse_context, observe
from pydantic import BaseModel
from tenacity import AsyncRetrying, RetryError, stop_after_attempt, wait_exponential

from agentlens.cache import TaskCache
from agentlens.dataset import Dataset, Row
from agentlens.hooks import Hook, Hooks
from agentlens.log import create_run_log
from agentlens.provider import Message, Provider
from agentlens.serialization import serialize_task_input, serialize_task_output
from agentlens.utils import create_path

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=BaseModel)
D = TypeVar("D", bound=Dataset)
P = ParamSpec("P")
R = TypeVar("R")
RunT = TypeVar("RunT")


class AI:
    langfuse: Langfuse | None
    message: Type[Message] = Message
    _dataset_dir: Path
    _run_dir: Path
    _providers: dict[str, Provider]

    def __init__(
        self,
        *,
        dataset_dir: Path | str,
        run_dir: Path | str,
        langfuse: Langfuse | None = None,
        providers: Sequence[Provider] = (),
    ):
        self.langfuse = langfuse
        self._dataset_dir = create_path(dataset_dir)
        self._run_dir = create_path(run_dir)
        self._providers = {provider.name: provider for provider in providers}

    def _create_messages(
        self,
        *,
        messages: list[Message] | None = None,
        system: str | None = None,
        prompt: str | None = None,
        dedent: bool = True,
    ) -> list[Message]:
        # check for invalid combinations
        if messages and (system or prompt):
            raise ValueError("Cannot specify both 'messages' and 'system'/'prompt'")

        # create messages if passed prompts
        if not messages:
            messages = []
            if system:
                messages.append(Message.system(system))
            if prompt:
                messages.append(Message.user(prompt))

        # apply dedent if needed
        return messages if not dedent else [m.dedent() for m in messages]

    def _parse_model_id(self, model_id: str) -> tuple[str, str]:
        """Transforms 'provider:model' -> (provider, model)"""
        try:
            provider_name, model_name = model_id.split(":", 1)
            return provider_name, model_name
        except ValueError:
            raise ValueError(
                f"Invalid model identifier '{model_id}'. " f"Expected format: 'provider:model'"
            )

    def _get_provider(self, model_id: str) -> tuple[Provider, str]:
        """Get the provider and parsed model name for a given model identifier"""
        provider_name, model_name = self._parse_model_id(model_id)

        provider = self._providers.get(provider_name)
        if not provider:
            raise ValueError(
                f"Provider '{provider_name}' not configured in AI manager. "
                f"Available providers: {list(self._providers.keys())}"
            )

        return provider, model_name

    async def _generate(
        self,
        generator: Callable,
        *,
        model: str,
        semaphore: asyncio.Semaphore,
        messages: list[Message] | None,
        system: str | None,
        prompt: str | None,
        dedent: bool,
        max_retries: int,
        **kwargs,
    ) -> Any:
        collected_messages = self._create_messages(
            messages=messages,
            system=system,
            prompt=prompt,
            dedent=dedent,
        )
        try:
            async for attempt in AsyncRetrying(
                stop=stop_after_attempt(max_retries),
                wait=wait_exponential(multiplier=1, min=1, max=10),
                reraise=True,
            ):
                with attempt:
                    try:
                        async with semaphore:
                            await asyncio.sleep(random.uniform(0, 0.1))
                            return await generator(
                                model=model,
                                messages=collected_messages,
                                **kwargs,
                            )
                    except Exception as e:
                        logger.debug(
                            f"Retry ({attempt.retry_state.attempt_number} of {max_retries}): {e}"
                        )
                        raise e
        except RetryError as e:
            logger.debug(f"Failed after {max_retries} attempts: {e}")
            raise e

    async def generate_text(
        self,
        *,
        model: str,
        messages: list[Message] | None = None,
        system: str | None = None,
        prompt: str | None = None,
        dedent: bool = True,
        max_retries: int = 3,
        **kwargs,
    ) -> str:
        provider, model_name = self._get_provider(model)
        return await self._generate(
            provider.generate_text,
            model=model_name,
            semaphore=provider.get_semaphore(model_name),
            messages=messages,
            system=system,
            prompt=prompt,
            dedent=dedent,
            max_retries=max_retries,
            **kwargs,
        )

    async def generate_object(
        self,
        *,
        model: str,
        type: Type[T],
        messages: list[Message] | None = None,
        system: str | None = None,
        prompt: str | None = None,
        dedent: bool = True,
        max_retries: int = 3,
        **kwargs,
    ) -> T:
        provider, model_name = self._get_provider(model)
        return await self._generate(
            provider.generate_object,
            model=model_name,
            semaphore=provider.get_semaphore(model_name),
            type=type,
            messages=messages,
            system=system,
            prompt=prompt,
            dedent=dedent,
            max_retries=max_retries,
            **kwargs,
        )

    def run(
        self,
        *,
        main: Callable[[Any], Awaitable[RunT]],
        hooks: list[Callable[[Row], Hook]] | None = None,
        dataset: Dataset,
    ) -> list[RunT]:
        # sketchy -- should only be used in evals
        nest_asyncio.apply()

        async def run_with_hooks(row: Row):
            _hooks = [hook_factory(row) for hook_factory in (hooks or [])]
            with Hooks.create(_hooks):
                return await main(row)

        @observe()
        async def run_all():
            url = langfuse_context.get_current_trace_url()
            if url:
                print(f"View trace: {url}")
            tasks = [run_with_hooks(row) for row in dataset]
            return await asyncio.gather(*tasks)

        with TaskCache.enable(self._dataset_dir / "cache"):
            with create_run_log(self._run_dir) as log:
                return asyncio.run(run_all(langfuse_observation_id=log.run_id))

    def hook(self, target_func: Callable, **kwargs) -> Callable[[Callable], Callable[[Row], Hook]]:
        def decorator(cb: Callable) -> Callable[[Row], Hook]:
            @wraps(cb)
            def wrapper(row: Row) -> Hook:
                return Hook(cb, target_func, row, **kwargs)

            return wrapper

        return decorator

    def score(
        self,
        name: str,
        value: Any,
        comment: str | None = None,
    ):
        if self.langfuse:
            self.langfuse.score(
                name=name,
                value=value,
                trace_id=langfuse_context.get_current_trace_id(),
                observation_id=langfuse_context.get_current_observation_id(),
                comment=comment,
            )

    def task(
        self,
        cache: bool = False,
        capture_input: bool = True,
        capture_output: bool = True,
    ) -> Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]:
        def decorator(func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
            task_name = func.__name__

            # conditionally cache
            if cache:
                func = TaskCache.cached(func)

            @wraps(func)
            @observe(capture_input=capture_input, capture_output=capture_output)
            async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                # execute function
                output = await func(*args, **kwargs)

                # run any hooks
                if (hooks := Hooks.get()):
                    for hook in hooks[task_name]:
                        hook(output, *args, **kwargs)

                # log to Langfuse
                if self.langfuse:
                    langfuse_context.update_current_observation(
                        name=task_name,
                        input=serialize_task_input(args, kwargs) if capture_input else None,
                        output=serialize_task_output(output) if capture_output else None,
                    )

                return output

            return wrapper

        return decorator

    def dataset(self, name: str) -> Callable[[Type[D]], Type[D]]:
        def decorator(cls: Type[D]) -> Type[D]:
            cls.name = name
            cls.dataset_dir = self._dataset_dir
            return cls

        return decorator



from typing import Type, TypeVar

from langfuse.openai import AsyncOpenAI
from pydantic import BaseModel

from agentlens.provider import Message, Provider

T = TypeVar("T", bound=BaseModel)


class OpenAIProvider(Provider):
    def __init__(
        self,
        api_key: str | None = None,
        max_connections: dict[str, int] = {"DEFAULT": 10},
    ):
        super().__init__(name="openai", max_connections=max_connections)
        self.client = AsyncOpenAI(api_key=api_key)

    async def generate_text(
        self,
        *,
        model: str,
        messages: list[Message],
        **kwargs,
    ) -> str:
        completion = await self.client.chat.completions.create(
            model=model,
            messages=[{"role": m.role, "content": m.content} for m in messages],
            **kwargs,
        )
        assert completion.choices[0].message.content is not None
        return completion.choices[0].message.content

    async def generate_object(
        self,
        *,
        model: str,
        messages: list[Message],
        type: Type[T],
        **kwargs,
    ) -> T:
        completion = await self.client.beta.chat.completions.parse(
            model=model,
            messages=[message.model_dump() for message in messages],
            response_format=type,
            **kwargs,
        )
        assert completion.choices[0].message.parsed is not None
        return completion.choices[0].message.parsed
